#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libgen.h>
#include <errno.h>
#include <sys/socket.h>
#include <linux/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <limits.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/wait.h>

#include "criu/criu-plugin.h"
#include "criu/criu-log.h"

#define BUF_SIZE 256

char *utc_date(char *buf, int len)
{
	time_t rawtime;
	struct tm *ptm;

	rawtime = time(NULL);
	if (rawtime == -1) {
		pr_warn("log_dumper: The time() function failed");
		return NULL;
	}

	ptm = localtime(&rawtime);
	if (ptm == NULL) {
		pr_warn("log_dumper: The localtime() function failed");
		return NULL;
	}

	if (strftime(buf, len, "%Y-%m-%d", ptm) == 0) {
		pr_warn("log_dumper: The strftime() function failed");
		return NULL;
	}

	pr_info(buf);
	return buf;
}

char *strstrip(const char *str)
{
	int left = 0, len = strlen(str), right;
	char *res;

	while (left < len && isspace(str[left]))
		left++;

	if (left == len) {
		char *ret = malloc(1);
		if (ret == NULL)
			return NULL;
		ret[0] = '\0';
		return ret;
	}

	right = len - 1;
	while (right >= 0 && isspace(str[right]))
		right--;

	res = malloc(right - left + 2);
	if (res == NULL)
		return NULL;

	strncpy(res, &str[left], right - left + 1);
	res[right - left + 1] = '\0';

	return res;
}

int check_report_enabled(char const *desired_name, char const *status)
{
	char *key = (char *)malloc(BUF_SIZE);
	char *value = (char *)malloc(BUF_SIZE);
	static const char filename[] = "/etc/vz/reporter-vz-prlrep.conf";
	FILE *file;

	file = fopen(filename, "r");
	if (file != NULL) {
		char line[128];

		while (fgets(line, sizeof(line), file) != NULL) {
			if (2 == sscanf(line, "%127[^=]=%127[^\n]%*c", key, value)) {
				char *stripped_key = strstrip(key);
				char *stripped_value = strstrip(value);

				if (!strcasecmp(stripped_key, desired_name) && !strcasecmp(stripped_value, status)) {
					fclose(file);
					free(stripped_key);
					free(stripped_value);
					free(key);
					free(value);
					return 1;
				}
				free(stripped_key);
				free(stripped_value);
			}
		}
		fclose(file);
	}
	free(key);
	free(value);
	return 0;
}

static inline bool dir_dots(const struct dirent *de)
{
	return !strcmp(de->d_name, ".") || !strcmp(de->d_name, "..");
}

static int cloexec_fds(void)
{
	DIR *dir;
	struct dirent *de;
	int fd;

	dir = opendir("/proc/self/fd");
	if (dir == NULL) {
		pr_warn("log_dumper: Can't open /proc/self/fd\n");
		return -1;
	}

	while ((de = readdir(dir))) {
		if (dir_dots(de))
			continue;

		if (sscanf(de->d_name, "%d", &fd) != 1) {
			pr_warn("log_dumper: Can't parse %s\n", de->d_name);
			closedir(dir);
			return -1;
		}

		if (fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC)) {
			pr_warn("log_dumper: Can't make %d FD_CLOEXEC\n", fd);
			closedir(dir);
			return -1;
		}
	}
	closedir(dir);

	return 0;
}

/*
 * !IMPORTANT!
 * Don't ever call exit from plugin function, it would lead to criu caller
 * exit. Note: exit from forked child is ok, and children should always either
 * exec or exit.
 */
void log_dumper(int stage, int ret)
{
	char dest[BUF_SIZE + 1] = "/var/spool/abrt-criu/criu-";
	char *args[] = { "/usr/libexec/component_maker", NULL };
	char criu_workdir[PATH_MAX];
	pid_t pid;
	char *off;
	int len;

	if (ret) {
		off = dest + strlen(dest);
		len = BUF_SIZE - strlen(dest);

		if (!utc_date(off, len))
			return;

		if (criu_get_work_dir_path(criu_workdir, sizeof(criu_workdir))) {
			pr_warn("log_dumper: The criu_get_work_dir_path() function failed");
			return;
		}

		setenv("DUMP_DIR", dest, 0);
		setenv("CRIU_WORKDIR", criu_workdir, 0);

		pr_info("migration failed, running log collector\n");
		pr_info("Criu work directory %s\n", getenv("CRIU_WORKDIR"));
		pr_info("Dump directory %s\n", getenv("DUMP_DIR"));

		pid = fork();
		if (pid < 0) {
			pr_warn("log_dumper: The fork() function failed");
			return;
		} else if (pid == 0) {
			/* child */
			if (check_report_enabled("criuReportEnabled", "false")) {
				pr_info("Report is disabled\n");
				exit(0);
			}

			/*
			 * Fork second child so that after first child dies
			 * second child is fully daemonized
			 */
			pid = fork();
			if (pid < 0) {
				pr_warn("log_dumper: The fork() function failed");
				exit(0);
			} else if (pid == 0) {
				/* second child */
				if (cloexec_fds()) {
					pr_warn("Can't setup fds to be closed on exec\n");
					exit(0);
				}

				if (setsid() == -1) {
					pr_warn("Can't create new session\n");
					exit(0);
				}

				/* just execute log component_maker and don't hunt exit code */
				execvp(args[0], args);
				exit(0);
			}
			exit(0);
		} else {
			waitpid(pid, NULL, 0);
		}
	}
}

CR_PLUGIN_REGISTER("logdumper", cr_plugin_dummy_init, log_dumper)
